/**************************************************************************
***    
*** Copyright (c) 2000-2004 Regents of the University of Michigan,
***               Saurabh N. Adya, Hayward H. Chan and Igor L. Markov
***
***  Contact author(s): sadya@eecs.umich.edu, imarkov@umich.edu
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/
















#ifndef DB_H
#define DB_H

#include <vector>
#include <algorithm>
#include <math.h>
#include <stdlib.h>
#include "Nets.h"
#include "Nodes.h"


namespace parquetfp
{
   class DB
   {
   protected:
      Nodes* _nodes;
      Nets* _nets;
  
      Nodes* _nodesBestCopy;

      double _area;
      bool _initArea;

      double _rowHeight;   //rowHeight to snap the soln to
      double _siteSpacing; //site spacing to snap the soln to

   public:
      bool successAR;
      DB(char* baseName);
      DB(DB * db, vector<int>& subBlocksIndices, Point& dbLoc, double reqdAR);
      DB(void);

      //makes a copy of db2. compresses multiple 2-pin nets
      //between modules regardless of pin-offset. use only when pin-offsets
      //are inconsequential
      DB(DB& db2, bool compressDB=false);

      ~DB();

      DB& operator=(DB& db2);
      void clean(void);

      unsigned getNumNodes(void) const;
      Nodes* getNodes(void);
      Nets* getNets(void);

      vector<double> getNodeWidths() const;
      vector<double> getNodeHeights() const;
      vector<double> getXLocs() const;
      vector<double> getYLocs() const;

      Point getBottomLeftCorner() const;
      Point getTopRightCorner() const;

      double getNodesArea(void);
      double getRowHeight(void) const;
      double getSiteSpacing(void) const;
      void setRowHeight(double rowHeight);
      void setSiteSpacing(double siteSpacing);

      // slim:       xloc, yloc, orient, width, height (but not the orig's)
      // location:   xloc, yloc
      // dimensions: width, height (but not the orig's)
      // TRUE ~ succeed in updating, FALSE otherwise
      inline bool updateNodeSlim(int index, const Node& newNode);
      inline bool updateNodeLocation(int index,
                                     double xloc, double yloc);
      inline bool updateNodeDimensions(int index,
                                       double width, double height);
      
      void updatePlacement(const vector<double>& xloc,
                           const vector<double>& yloc);
      void initPlacement(const Point& loc);      
      void updateSlacks(const vector<double>& xSlack,
                        const vector<double>& ySlack);
      
      void plot(char* fileName, double area, double whitespace, double aspectRatio,
                double time, double HPWL, bool plotSlacks, bool plotNets, bool plotNames);

      void saveCapo(char *baseFileName, double reqdAR=1, double reqdWS=30);  //save the data in Capo format
      void saveCapoNets(char* baseFileName);

      void save(char* baseFileName); //save data in floorplan format
      void saveNets(char* baseFileName);
      void saveWts(char* baseFileName);

      void saveBestCopyPl(char* baseFileName);
      void saveInBestCopy(void);

      double evalHPWL(bool useWts=true) const;  //assumes that placement is updated
      //assumes that placement is updated & terminals to be moved to floorplan edges
      double evalHPWL(double xSize, double ySize, bool useWts=true) const;
      double evalArea(void);  //assumes that placement is updated
      double getXSize(void);
      double getYSize(void);
      double getAvgHeight(void) const;

      // optimize the BL-corner of the design
      void shiftOptimizeDesign(double outlineWidth,
                               double outlineHeight); // deduce BL-corner from DB
      void shiftOptimizeDesign(const Point& bottomLeft,
                               const Point& topRight); 
      
      void shiftDesign(Point& offset); //shift the entire placement by an offset

      void expandDesign(double maxWidth, double maxHeight);
      //expand entire design to occupy the new dimensions. only locations of blocks
      //are altered
  
      //marks all nodes with height > avgHeight as macros
      void markTallNodesAsMacros(double maxHeight);

      //reduce the core cells area of design excluding macros to maxWS whitespace
      //assumes macros are marked using DB::markTallNodesAsMacros()
      void reduceCoreCellsArea(double layoutArea, double maxWS);

      //this function gets the dimension of the FP only considering the macros
      double getXSizeWMacroOnly();
      double getYSizeWMacroOnly();

      double getXMaxWMacroOnly();
      double getYMaxWMacroOnly();

   private:
      // help function of shiftOptimizeDesign()
      double getOptimalRangeStart(bool isHorizontal);
   };

   // ---------------
   // IMPLEMENTATIONS
   // ---------------
   bool DB::updateNodeSlim(int index,
                           const Node& newNode)
   {
      if (index < int(_nodes->getNumNodes()))
      {
         Node& oldNode = _nodes->getNode(index);
         
         oldNode.putX(newNode.getX());
         oldNode.putY(newNode.getY());
         oldNode.changeOrient(newNode.getOrient(), *_nets);
         oldNode.putWidth(newNode.getWidth());
         oldNode.putHeight(newNode.getHeight());
         return true;
      }
      else
         return false;
   }
   // -----------------------------------------------------
   bool DB::updateNodeLocation(int index,
                               double xloc, double yloc)
   {
      if (index < int(_nodes->getNumNodes()))
      {
         Node& oldNode = _nodes->getNode(index);
         oldNode.putX(xloc);
         oldNode.putY(yloc);
         return true;
      }
      else
         return false;
   }
   // -----------------------------------------------------
   bool DB::updateNodeDimensions(int index,
                                 double width, double height)
   {
      if (index <int( _nodes->getNumNodes()))
      {
         Node& oldNode = _nodes->getNode(index);

         oldNode.putWidth(width);
         oldNode.putHeight(height);
         return true;
      }
      else
         return false;
   }
   // -----------------------------------------------------
}
// using namespace parquetfp;

#endif 
